from django.apps import apps
from django.contrib.admin.views.decorators import staff_member_required
from django.http import (
    HttpResponse,
    HttpResponseBadRequest,
    HttpResponseNotAllowed,
)
from django.shortcuts import redirect

from coffee.util.decorators import has_sufficient_params
from coffee.util.functions import (
    get_delete_button_html,
    render_model_form_html,
)


@staff_member_required
@has_sufficient_params
def get_model_form(request):
    app_name = request.GET.get("app_name")
    model_name = request.GET.get("model_name")
    pk = request.GET.get("pk")

    instance = None

    if pk:
        try:
            cls = apps.get_model(app_name, model_name)
            instance = cls.objects.get(pk=pk)
        except cls.DoesNotExist:
            exception = f"pk '{pk}' for '{app_name}.{model_name}' not found"
            return HttpResponseBadRequest(exception)

    html = "<div class='Wrapper'>\n"
    html += "<div class='Header'>\n"
    if instance:
        html += get_delete_button_html(request, app_name, model_name) % instance.pk
    html += "</div>"
    html += render_model_form_html(request, app_name, model_name, instance)
    html += "</div>"

    return HttpResponse(html)


@staff_member_required
@has_sufficient_params
def post_model_form(request):
    app_name = request.GET.get("app_name")
    model_name = request.GET.get("model_name")

    cls = apps.get_model(app_name, model_name)
    pk_field = cls._meta.pk.name

    if request.method != "POST":
        return HttpResponseNotAllowed(
            f"'{request.method} is not valid for this endpoint."
        )

    data = {
        k: None if v == "" else v
        for k, v in request.POST.copy().items()
        if k != "csrfmiddlewaretoken"
    }
    pk = data.get(pk_field)

    if not pk:
        instance, _ = cls.objects.get_or_create(**data)
        pk = instance.pk

    if pk:
        instance = cls.objects.get(pk=pk)
        for key, value in data.items():
            setattr(instance, key, value)
        instance.save()

    url = f"/coffee/form/?app_name={app_name}&model_name={model_name}&pk={pk}"

    return redirect(url)


@staff_member_required
@has_sufficient_params
def delete_model_instance(request, pk=None):
    app_name = request.GET.get("app_name")
    model_name = request.GET.get("model_name")

    cls = apps.get_model(app_name, model_name)

    if request.method != "POST":
        return HttpResponseNotAllowed(
            f"'{request.method} is not valid for this endpoint."
        )

    if pk:
        instance = cls.objects.get(pk=pk)
        instance.delete()

    url = f"/coffee/form/?app_name={app_name}&model_name={model_name}"

    return redirect(url)



TYPESCRIPT_TRANSLATION = {
    "AutoField": "number",
    "BigAutoField": "number",
    "BigIntegerField": "number",
    "BinaryField": "number",
    "BooleanField": "boolean",
    "CharField": "text",
    "CommaSeparatedIntegerField": "number",
    "DateField": "Date",
    "DateTimeField": "Date",
    "DecimalField": "number",
    "DurationField": "number",
    "EmailField": "text",
    "FilePathField": "text",
    "FloatField": "number",
    "GenericIPAddressField": "text",
    "IPAddressField": "text",
    "IntegerField": "number",
    "NullBooleanField": "boolean | null",
    "PositiveBigIntegerField": "number",
    "PositiveIntegerField": "number",
    "PositiveSmallIntegerField": "number",
    "SlugField": "string",
    "SmallAutoField": "number",
    "SmallIntegerField": "number",
    "TextField": "string",
    "TimeField": "string",
    "URLField": "string",
    "UUIDField": "string",
    "ForeignKey": "number",
    "OneToOneField": "number",
}

FORM_TRANSLATION = {
    "AutoField": "<input value='%s' type='number' name='%s' />",
    "BigAutoField": "<input value='%s' type='number' name='%s' />",
    "BigIntegerField": "<input value='%s' type='number' name='%s' />",
    "BinaryField": "<input value='%s' type='number' name='%s' />",
    "BooleanField": "<input value='%s' type='checkbox' name='%s' />",
    "CharField": "<input value='%s' name='%s'/>",
    "CommaSeparatedIntegerField": ("<input value='%s' type='number' name='%s' />"),
    "DateField": "<input value='%s' type='date' name='%s'/>",
    "DateTimeField": "<input value='%s' type='date' name='%s'/>",
    "DecimalField": "<input value='%s' type='number' name='%s' />",
    "DurationField": "<input value='%s' type='number' name='%s' />",
    "EmailField": "<input value='%s' name='%s'/>",
    "FilePathField": "<input value='%s' name='%s'/>",
    "FloatField": "<input value='%s' type='number' name='%s' />",
    "GenericIPAddressField": "<input value='%s' name='%s'/>",
    "IPAddressField": "<input value='%s' name='%s'/>",
    "IntegerField": "<input value='%s' type='number' name='%s' />",
    "NullBooleanField": "<input value='%s' type='checkbox' name='%s' />",
    "PositiveBigIntegerField": "<input value='%s' type='number' name='%s' />",
    "PositiveIntegerField": "<input value='%s' type='number' name='%s' />",
    "PositiveSmallIntegerField": ("<input value='%s' type='number' name='%s' />"),
    "SlugField": "<input value='%s' name='%s' />",
    "SmallAutoField": "<input value='%s' type='number' name='%s' />",
    "SmallIntegerField": "<input value='%s' type='number' name='%s' />",
    "TextField": "<input value='%s' type='text' name='%s' />",
    "TimeField": "<input value='%s' name='%s' />",
    "URLField": "<input value='%s' name='%s' />",
    "UUIDField": "<input value='%s' name='%s' />",
    "ForeignKey": "<input value='%s' type='number' name='%s' />",
    "OneToOneField": "<input value='%s' type='number' name='%s' />",
}



import re

from importlib import import_module
from inspect import getmembers, isclass

from coffee.exceptions import DefinitionNotFound, FieldNotSupported
from .definitions import FORM_TRANSLATION, TYPESCRIPT_TRANSLATION

from django.middleware.csrf import get_token


def get_typescript_type(db_type):
    if db_type not in TYPESCRIPT_TRANSLATION:
        raise FieldNotSupported(db_type)
    return TYPESCRIPT_TRANSLATION[db_type]


def to_camel_case(text):
    first, *others = text.split("_")
    return "".join([first.lower(), *map(str.title, others)])


def to_snake_case(text):
    text = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", text)
    text = re.sub("__([A-Z])", r"_\1", text)
    text = re.sub("([a-z0-9])([A-Z])", r"\1_\2", text)
    return text.lower()


def get_model_definitions(app_name):
    module = import_module(f"{app_name}.models")
    models = getmembers(module, predicate=isclass)

    definitions = {}

    for model_name, model in models:
        if model_name not in definitions:
            definitions[model_name] = {}

        for field in model._meta.fields:
            related_model = None

            if field.related_model:
                related_model = field.related_model.__name__

            definitions[model_name][field.name] = {
                "related_model": related_model,
                "primary_key": field.primary_key,
                "db_type": field.get_internal_type(),
                "nullable": field.null,
            }

    return definitions


def get_form_item_input_html(db_type):
    if db_type not in FORM_TRANSLATION:
        raise FieldNotSupported()

    return f"{FORM_TRANSLATION[db_type]}\n"


def get_form_item_label_html():
    return "<label for='%s'>%s</label>\n"


def get_form_item_row_html(name, db_type, value=None):
    html = '<div class="FormItem">\n'
    html += get_form_item_label_html() % (name, name)
    html += get_form_item_input_html(db_type=db_type) % (value, name)
    html += "</div>\n"
    return html


def get_form_submit_button_html():
    html = '<div class="FormItem">\n'
    html += "<button type='submit' class='FormButton'>Submit</button>"
    html += "</div>\n"
    return html


def get_csrf_token_html():
    return '<input type="hidden" name="csrfmiddlewaretoken" value="%s" />'


def get_delete_button_html(request, app_name, model_name):
    csrf_token = get_token(request)
    url = f"/coffee/delete/%s/?app_name={app_name}&model_name={model_name}"
    html = f"<form method='POST' action='{url}' >"
    html += get_csrf_token_html() % (csrf_token)
    html += "<button type='submit'>Delete</button>\n"
    html += "</form>\n"
    return html


def render_model_form_html(request, app_name, model_name, instance):
    definitions = get_model_definitions(app_name)

    csrf_token = get_token(request)

    if model_name not in definitions:
        raise DefinitionNotFound(model_name)

    definition = definitions[model_name]

    html = "<form method='POST' class='Form' action='/coffee/form/submit/"
    html += f"?app_name={app_name}&model_name={model_name}' >\n"

    for name, field in definition.items():
        db_type = field.get("db_type")
        html += get_form_item_row_html(
            name=name, db_type=db_type, value=getattr(instance, name, None)
        )

    html += get_form_submit_button_html()
    html += get_csrf_token_html() % (csrf_token)

    html += "</form>\n"

    return html



class FieldNotSupported(Exception):
    def __init__(self, field_name=None, *args, **kwargs):
        exception = f"Could not convert field '{field_name}'"
        return super().__init__(exception, *args, **kwargs)


class DefinitionNotFound(Exception):
    def __init__(self, model_name=None, *args, **kwargs):
        exception = f"Could not find model '{model_name}'"
        return super().__init__(exception, *args, **kwargs)



from django.urls import path

from coffee.views import delete_model_instance, get_model_form, post_model_form


urlpatterns = [
    path("form/", get_model_form, name="form"),
    path("form/submit/", post_model_form, name="submit_form"),
    path("delete/<int:pk>/", delete_model_instance, name="delete"),
]
